/* --------------------------------------------------------------------- */
/*					Blob Manager Low-level Drawing Routines				 */
/* --------------------------------------------------------------------- */

# include	<GestaltEqu.h>
# include	<Palettes.h>		/* for GetGray() */

# include	"TransSkel.h"

# include	"BlobMgr.h"


/*
 * Blob dimming parameters.  dimPat and dimMode are the pattern and
 * drawing mode used to dim out blobs.  The initial pattern is gray.
 */

static Pattern	dimPat = { 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55 };
static short	dimMode = patBic;

static RgnHandle	oldClipRgn;


/*
 * Set the clipping to the given region of the given blob.  Return
 * a handle to the region, or nil if the region is empty and doesn't
 * need to be drawn.  The handle should not be disposed of by the
 * caller - it's a copy of the handle only, not a copy of the region
 * itself.  If the handle returned is not nil, call RestoreDrawClip
 * after doing the drawing to restore the clipping properly.
 */

static RgnHandle
SetDrawClip (BlobHandle b, short partCode)
{
RgnHandle	drawRgn, newClipRgn;

	drawRgn = (partCode == inDragBlob ? (**b).dragRgn : (**b).statRgn);
	if (drawRgn == (RgnHandle) nil || EmptyRgn (drawRgn))
		return ((RgnHandle) nil);		/* don't bother if fail here! */

	oldClipRgn = NewRgn ();
	newClipRgn = NewRgn ();
	GetClip (oldClipRgn);
	SectRgn (drawRgn, oldClipRgn, newClipRgn);
	SetClip (newClipRgn);
	DisposeRgn (newClipRgn);
	return (drawRgn);
}


static void
RestoreDrawClip (void)
{
	SetClip (oldClipRgn);				/* restore original clipping */
	DisposeRgn (oldClipRgn);
}


/*
 * Set or get the pattern and drawing mode to be used to dim blob
 * regions.  Default is to dither a gray pattern over the region.
 */

pascal void
SetBDimInfo (Pattern p, short mode)
{
	BlockMove (p, &dimPat, sizeof (Pattern));
	dimMode = mode;
}


pascal void
GetBDimInfo (Pattern *p, short *mode)
{
	BlockMove (&dimPat, p, sizeof (Pattern));
	*mode = dimMode;
}


/*
 * Dim the given region.
 */

static void
DimRgn (RgnHandle rgn)
{
PenState	ps;

	if (rgn != (RgnHandle) nil)
	{
		GetPenState (&ps);			/* save current pen state */
		PenPat (dimPat);
		PenMode (dimMode);
		PaintRgn (rgn);
		SetPenState (&ps);			/* restore pen state */
	}
}


/*
 * Dim the region of a blob indicated by the part code by dithering
 * a gray pattern over it.  The affected region is assumed to be
 * drawn already.
 */

pascal void
DimBlobRgn (BlobHandle b, short partCode)
{
	if (partCode == inStatBlob || partCode == inFullBlob)
		DimRgn ((**b).statRgn);
	if (partCode == inDragBlob || partCode == inFullBlob)
		DimRgn ((**b).dragRgn);
}


/*
 * Set up for drawing a dimmed blob region.  The blob and the region are indicated
 * by b and partCode.  The return value is true if the function was able to change
 * then pen for drawing in gray, false otherwise.
 *
 * After calling this, draw the blob region, then call EndBlobDimDraw() to restore
 * the pen color.  EndBlobDimDraw() will restore the pen color if it was changed
 * to gray by BeginBlobDimDraw(), otherwise it will dither the blob region with
 * patterned gray, which allows dimming even when true gray drawing isn't available.
 */

static Boolean	canGrayDraw;
static RGBColor	savedFgColor;
static BlobHandle	savedBlob;
static short		savedPartCode;

pascal Boolean
BeginBlobDimDraw (BlobHandle b, short partCode)
{
GrafPtr	port;
Rect	r;
RGBColor	fgColor, bgColor;
GDHandle	gDev;

	savedBlob = b;
	savedPartCode = partCode;

	/*
	 * Try to get RGB gray value appropriate for the device on which
	 * the region is displayed if have color GrafPort.
	 */
	canGrayDraw = false;
	GetPort (&port);
	if (((CGrafPtr) port)->portVersion & 0xc000)	/* is it a Color GrafPort? */
	{
		r = (partCode == inStatBlob ? BStatBox (b) : BDragBox (b));
		/* convert rect to global coordinates */
		LocalToGlobal (&topLeft (r));
		LocalToGlobal (&botRight (r));
		(void) SkelGetRectDevice (&r, &gDev, (Rect *) nil, (Boolean *) nil);
		if (gDev != (GDHandle) nil)
		{
			GetBackColor (&bgColor);
			GetForeColor (&fgColor);
			savedFgColor = fgColor;
			canGrayDraw = GetGray (gDev, &bgColor, &fgColor);
		}
	}

	if (canGrayDraw)
		RGBForeColor (&fgColor);
	return (canGrayDraw);
}


/*
 * End dim drawing.  If could not draw in gray, then dither gray pattern over
 * affected region.  Otherwise undo changes that were made to start gray
 * drawing.
 */

pascal void
EndBlobDimDraw (void)
{
	if (!canGrayDraw)
		DimBlobRgn (savedBlob, savedPartCode);
	else
		RGBForeColor (&savedFgColor);	/* restore foreground color */
}


/*
 * Given a picture and a rectangle specifying a picture frame,
 * DrawBlobPic draws the picture in the given part of the given blob,
 * accounting for size and location differences.  Any portion of srcRect
 * lying outside of the picture's frame will not be drawn.
 * The partCode must be inDragBlob or inFullBlob.
 *
 * The drawing is clipped to the region to be drawn
 * (respecting current clipping).
 */

static void
DrawBlobPic (BlobHandle b, short partCode, PicHandle p, Rect *srcRect)
{
RgnHandle	rgn;
Rect		dstRect, pFrame;

	rgn = SetDrawClip (b, partCode);
	if (rgn != (RgnHandle) nil && p != (PicHandle) nil)
	{
		pFrame = (**p).picFrame;			/* map frame to correct size */
		dstRect = (**rgn).rgnBBox;			/* and location in relation */
		MapRect (&pFrame, srcRect, &dstRect);	/* to region to be drawn */
		DrawPicture (p, &pFrame);
		if (GetBDrawMode (b, partCode) == dimDraw)	/* region is dimmed */
			DimRgn (rgn);
	}
	if (rgn != (RgnHandle) nil)
		RestoreDrawClip ();				/* restore original clipping */
}


/*
 * Draw a blob using a drawing procedure.  bDst is the blob to be drawn,
 * partCode specifies the region of bDst to draw (either inDragBlob or
 * inStatBlob).  bSrc is the blob whose region is to be drawn in the
 * specified region of bDst, and proc is the drawing procedure to use.
 * bSrc is different than bDst if partCode is inDragBlob and bDst has
 * a glob.
 *
 * For dimmed blobs:
 * if the bProcManualDimMask bit is set in the source blob's flags, the
 * proc is assumed to do its own dimming "manually".  Otherwise, the pen
 * is changed to gray before calling the drawing proc and restored afterward.
 */

static void
CallBDrawProc (BlobHandle bDst, BlobHandle bSrc, BDrawProcPtr proc, short partCode)
{
RgnHandle	rgn;
Boolean		autoDim;

	rgn = SetDrawClip (bDst, partCode);
	if (rgn != (RgnHandle) nil && proc != (BDrawProcPtr) nil)
	{
		/* check whether to autodim the region and set the pen gray if so */
		autoDim = (GetBDrawMode (bDst, partCode) == dimDraw)
					&& !TestBlobFlags (bDst, bProcManualDimMask);
		if (autoDim)
			BeginBlobDimDraw (bDst, partCode);
		(*proc) (bDst, bSrc, partCode);
		if (autoDim)
			EndBlobDimDraw ();
	}
	if (rgn != (RgnHandle) nil)
		RestoreDrawClip ();				/* restore original clipping */
}


/*
 * DrawBlob draws the indicated part of the blob (the entire blob
 * if partCode is inFullBlob).
 *
 * The blob must be enabled.  If draw mode is dimDraw then draw
 * it dimmed.
 */


pascal void
DrawBlob (BlobHandle b, short partCode)
{
BlobHandle	b2;
PicHandle	pic;
Rect		srcFrame;

	if (BlobEnabled (b))
	{
		if ((partCode == inStatBlob) || (partCode == inFullBlob))
		{
			if (PicBlob (b))
			{
				srcFrame = (**b).statRect;
				/*srcFrame = BStatBox (b);*/
				DrawBlobPic (b, inStatBlob, (**b).bPicProc.bPic, &srcFrame);
			}
			else
				CallBDrawProc (b, b, (**b).bPicProc.bDrawProc, inStatBlob);
		}

		if ((partCode == inDragBlob) || (partCode == inFullBlob))
		{
		/*
		 * Set b2 to the blob that should have it's drag region drawn
		 * in b's drag region.  If b has no glob, b2 is b itself.
		 */
			if ((b2 = BGlob (b)) == (BlobHandle) nil)
				b2 = b;
			if (PicBlob (b2))
			{
				pic = (**b2).bPicProc.bPic;	/* picture to draw */
				srcFrame = (**b2).dragRect;	/* part of picture to draw */
				DrawBlobPic (b, inDragBlob, pic, &srcFrame);
			}
			else
				CallBDrawProc (b, b2, (**b2).bPicProc.bDrawProc, inDragBlob);
		}
	}
}


pascal void
DrawBlobSet (BlobSetHandle bSet)
{
	BlobLoopProc2 (&DrawBlob, bSet, inFullBlob);
}
